//---------------------------------------------------------------------------
#include <vcl.h>
#include <math.h>
#pragma hdrstop

#include "AlphaBlender.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TAlphaBlenderForm *AlphaBlenderForm;
//---------------------------------------------------------------------------
bool GrasomusAlphaBlend::isMMX()
{
  if (IsProcessorFeaturePresent(PF_MMX_INSTRUCTIONS_AVAILABLE)) return true;
  SYSTEM_INFO si;
  int nCPUFeatures=0;
  GetSystemInfo(&si);
  if (si.dwProcessorType != PROCESSOR_INTEL_386 && si.dwProcessorType != PROCESSOR_INTEL_486)
  {
    try
    {
      asm
      {
        push eax
        push ebx
        push ecx
        push edx

        mov eax,1
        mov nCPUFeatures,edx

        pop edx
        pop ecx
        pop ebx
        pop eax
      }
    }
    catch(...) // just to be sure...
    {
      return false;
    }
  }
  return (nCPUFeatures & 0x00800000) != 0;
}
//---------------------------------------------------------------------------
__fastcall TAlphaBlenderForm::TAlphaBlenderForm(TComponent* Owner)
        : TForm(Owner)
{
  ClientWidth=1;
  ClientHeight=1;
  Blending=true;
  Delay=3;
  g_isMMX = AlphaB.isMMX();
}
//---------------------------------------------------------------------------
__fastcall TAlphaBlenderForm::~TAlphaBlenderForm()
{
  if (frames)
    delete frames;
  if (backgnd)
    delete backgnd;
  if (image)
    delete image;
}
//---------------------------------------------------------------------------
void __fastcall TAlphaBlenderForm::FormShow(TObject *Sender)
{
  Error=false;

  if (g_isMMX)
  {
    NO_OF_PIXELS=PictureMMX->Canvas->ClipRect.Width()*PictureMMX->Canvas->ClipRect.Height();
    if (!image)
      image = new PixelInfo [NO_OF_PIXELS];
    else
    {
      delete image;
      image = new PixelInfo [NO_OF_PIXELS];
    }

    IMAGE_DIM_X=PictureMMX->Canvas->ClipRect.Width();
    IMAGE_DIM_Y=PictureMMX->Canvas->ClipRect.Height();

    unsigned RGB;
    for (unsigned j=0; j<IMAGE_DIM_Y; j++)
    {
      for (unsigned i=0; i<IMAGE_DIM_X; i++)
      {
        RGB = PictureMMX->Canvas->Pixels[i][j];
        image[(j*IMAGE_DIM_X)+i].r = GetB24Bit(RGB);
        image[(j*IMAGE_DIM_X)+i].g = GetG24Bit(RGB);
        image[(j*IMAGE_DIM_X)+i].b = GetR24Bit(RGB);
      }
    }
  }
  else
  {
    NO_OF_PIXELS=Picture->Canvas->ClipRect.Width()*Picture->Canvas->ClipRect.Height();
    if (!image)
      image = new PixelInfo [NO_OF_PIXELS];
    else
    {
      delete image;
      image = new PixelInfo [NO_OF_PIXELS];
    }

    IMAGE_DIM_X=Picture->Canvas->ClipRect.Width();
    IMAGE_DIM_Y=Picture->Canvas->ClipRect.Height();

    unsigned RGB;
    for (unsigned j=0; j<IMAGE_DIM_Y; j++)
    {
      for (unsigned i=0; i<IMAGE_DIM_X; i++)
      {
        RGB = Picture->Canvas->Pixels[i][j];
        image[(j*IMAGE_DIM_X)+i].r = GetB24Bit(RGB);
        image[(j*IMAGE_DIM_X)+i].g = GetG24Bit(RGB);
        image[(j*IMAGE_DIM_X)+i].b = GetR24Bit(RGB);
      }
    }
  }

  NO_OF_PIXELS=IMAGE_DIM_X*IMAGE_DIM_Y;
  AlphaB.NO_OF_PIXELS=NO_OF_PIXELS;

  if (!frames)
    frames = new unsigned [NO_OF_PIXELS];
  else
  {
    delete frames;
    frames = new unsigned [NO_OF_PIXELS];
  }
  if (!backgnd)
    backgnd = new PixelInfo[NO_OF_PIXELS];
  else
  {
    delete backgnd;
    backgnd = new PixelInfo[NO_OF_PIXELS];
  }

  SystemParametersInfo( SPI_SCREENSAVERRUNNING, 1, NULL, 0);
  oldcursor=GetCursor();
  ::SetCursor(NULL);

  // set frame count
  ShiftCount=8;

  // init GDI
  hDesktopDC=GetDC(NULL);
  hMemoryDC=CreateCompatibleDC(hDesktopDC);

  // window where we'll be drawing to
  AlphaBlenderFormDC=GetDC(Handle);

  // create tmp bitmap for frames
  hbm=CreateCompatibleBitmap(hDesktopDC, IMAGE_DIM_X, IMAGE_DIM_Y);
  GetObject(hbm, sizeof(BITMAP), &bm);
  hbmOld = SelectObject(hMemoryDC, hbm);

  BITSPERPIXEL = bm.bmBitsPixel;

  if (BITSPERPIXEL==8)
  {
    ::SetCursor(oldcursor);
    Error=true;
    return;
  }
  else
  if (BITSPERPIXEL==16) // desktop is in 16-bit color mode
  {
    // find out the pixel format of the display card
    unsigned short testpixel=(1<<10);
    SetBitmapBits(hbm, 2, &testpixel);
    COLORREF testcolor=GetPixel(hMemoryDC, 0, 0);
    if (!GetRValue(testcolor)) // 5-6-5 format
    {
      AlphaB.g_RShift=11;
      AlphaB.g_GShift=6;
    }
    else                       // 5-5-5 format
    {
      AlphaB.g_RShift=10;
      AlphaB.g_GShift=5;
    }
  }
  else
  if (BITSPERPIXEL==24)
  {
    // find out the pixel format of the display card
    char testpixel[]={1,2,3};
    SetBitmapBits(hbm, BITSPERPIXEL>>3, testpixel);
    COLORREF testcolor=GetPixel(hMemoryDC, 0, 0);
    if ((GetRValue(testcolor)!=3)||(GetGValue(testcolor)!=2)||(GetBValue(testcolor)!=1))
    {
      ShowMessage("Error in pixel format... a variation detected!");
      ShowMessage(IntToStr(GetRValue(testcolor)));
      ShowMessage(IntToStr(GetGValue(testcolor)));
      ShowMessage(IntToStr(GetBValue(testcolor)));
      Error=true;
      return;
    }
  }
  else
  if (BITSPERPIXEL==32)
  {
    // find out the pixel format of the display card
    char testpixel[]={1,2,3,4};
    SetBitmapBits(hbm, BITSPERPIXEL>>3, testpixel);
    COLORREF testcolor=GetPixel(hMemoryDC, 0, 0);
    if ((GetRValue(testcolor)!=3)||(GetGValue(testcolor)!=2)||(GetBValue(testcolor)!=1))
    {
      ShowMessage("Error in pixel format... a variation detected!");
      ShowMessage(IntToStr(GetRValue(testcolor)));
      ShowMessage(IntToStr(GetGValue(testcolor)));
      ShowMessage(IntToStr(GetBValue(testcolor)));
      Error=true;
      return;
    }
  }
  else
  {
    ::SetCursor(oldcursor);
    Error=true;
    return;
  }

  BYTESPERPIXEL=BITSPERPIXEL>>3;
  AlphaB.BYTESPERPIXEL=BYTESPERPIXEL;

  scr_x=Screen->Width/2-IMAGE_DIM_X/2;
  scr_y=Screen->Height/2-IMAGE_DIM_Y/2-1;

  RECT r;
  r.top=Screen->Height>>1;
  r.left=Screen->Width>>1;
  r.bottom=r.top+1;
  r.right=r.left+1;
  SetCursorPos(r.left, r.top);
  ClipCursor(&r);

  ::SetCursor(NULL);

  UpdateBackground();

  if (g_isMMX)
    AlphaB.AlphaBlendHardwareMMX(image, backgnd, frames, 0, ShiftCount);
  else
    AlphaB.AlphaBlendSoftwareMMX(image, backgnd, frames, 0, ShiftCount);
}
//---------------------------------------------------------------------------
void __fastcall TAlphaBlenderForm::UpdateBackground()
{
  // Get pixel by pixel info from desktop
  BitBlt(hMemoryDC, 0, 0, bm.bmWidth, bm.bmHeight, hDesktopDC, scr_x, scr_y, SRCCOPY);
  GetBitmapBits(hbm, NO_OF_PIXELS*BYTESPERPIXEL, frames);

  BYTE *tmpdest=(BYTE *)frames;

  for (unsigned i=0; i<NO_OF_PIXELS; i++)
  {
    if (BYTESPERPIXEL==2)
    {
      if (AlphaB.g_RShift==11)
      {
        backgnd[i].r = GetR16Bit(*((unsigned *)(tmpdest)));
        backgnd[i].g = GetG16Bit(*((unsigned *)(tmpdest)));
        backgnd[i].b = GetB16Bit(*((unsigned *)(tmpdest)));
      }
      else
      {
        backgnd[i].r = GetR15Bit(*((unsigned *)(tmpdest)));
        backgnd[i].g = GetG15Bit(*((unsigned *)(tmpdest)));
        backgnd[i].b = GetB15Bit(*((unsigned *)(tmpdest)));
      }
    }
    else
    if (BYTESPERPIXEL==3)
    {
      backgnd[i].r = GetR24Bit(*((unsigned *)(tmpdest)));
      backgnd[i].g = GetG24Bit(*((unsigned *)(tmpdest)));
      backgnd[i].b = GetB24Bit(*((unsigned *)(tmpdest)));
    }
    else
    if (BYTESPERPIXEL==4)
    {
      backgnd[i].r = GetR32Bit(*((unsigned *)(tmpdest)));
      backgnd[i].g = GetG32Bit(*((unsigned *)(tmpdest)));
      backgnd[i].b = GetB32Bit(*((unsigned *)(tmpdest)));
    }

    tmpdest+=BYTESPERPIXEL;
  }
}
//---------------------------------------------------------------------------
void __fastcall TAlphaBlenderForm::FormPaint(TObject *Sender)
{
  Top=Screen->Height/2-IMAGE_DIM_Y/2-1;
  Left=Screen->Width/2-IMAGE_DIM_X/2;

  ClientWidth=IMAGE_DIM_X;
  ClientHeight=IMAGE_DIM_Y;

  SetBitmapBits(hbm, NO_OF_PIXELS*AlphaB.BYTESPERPIXEL, frames);
  BitBlt(AlphaBlenderFormDC, 0, 0, bm.bmWidth, bm.bmHeight, hMemoryDC, 0, 0, SRCCOPY);

  unsigned start_cnt=GetTickCount();
  unsigned end_cnt;

  unsigned delay=Delay;

  for (unsigned short i=MinOpacity; i<MaxOpacity;)
  {
    if (BITSPERPIXEL==16)
    {
      if (g_isMMX)
        AlphaB.AlphaBlendHardwareMMX(image, backgnd, frames, i, ShiftCount);
      else
        AlphaB.AlphaBlendSoftwareMMX(image, backgnd, frames, i, ShiftCount);
    }
    else
      AlphaB.AlphaBlendBasic(image, backgnd, frames, i, ShiftCount);

    SetBitmapBits(hbm, NO_OF_PIXELS*AlphaB.BYTESPERPIXEL, frames);
    BitBlt(AlphaBlenderFormDC, 0, 0, bm.bmWidth, bm.bmHeight, hMemoryDC, 0, 0, SRCCOPY);

    do
    {
      end_cnt=GetTickCount();
    }while(end_cnt-start_cnt<delay);

    i+=(unsigned short)((end_cnt-start_cnt)/delay);

    start_cnt=end_cnt;
  }
  if (BITSPERPIXEL==16)
  {
    if (g_isMMX)
      AlphaB.AlphaBlendHardwareMMX(image, backgnd, frames, MaxOpacity, ShiftCount);
    else
      AlphaB.AlphaBlendSoftwareMMX(image, backgnd, frames, MaxOpacity, ShiftCount);
  }
  else
    AlphaB.AlphaBlendBasic(image, backgnd, frames, MaxOpacity, ShiftCount);
  SetBitmapBits(hbm, NO_OF_PIXELS*AlphaB.BYTESPERPIXEL, frames);
  BitBlt(AlphaBlenderFormDC, 0, 0, bm.bmWidth, bm.bmHeight, hMemoryDC, 0, 0, SRCCOPY);

  start_cnt=GetTickCount();
  do
  {
    end_cnt=GetTickCount();
  }while(end_cnt-start_cnt<PauseInterval);

  OnPaint=NULL;
  Blending=false;

}
//---------------------------------------------------------------------------
__inline void GrasomusAlphaBlend::AlphaBlendBasic(PixelInfo *image, PixelInfo *backgnd, unsigned *rendered_image, unsigned char src_coef, unsigned char ShiftCount)
{
  BYTE *tmpdest=(BYTE *)rendered_image;

  PixelInfo p;
  for (unsigned i=0; i<NO_OF_PIXELS; i++)
  {
    p.r=(unsigned char)(((src_coef*(image[i].r-backgnd[i].r))>>ShiftCount)+backgnd[i].r);
    p.g=(unsigned char)(((src_coef*(image[i].g-backgnd[i].g))>>ShiftCount)+backgnd[i].g);
    p.b=(unsigned char)(((src_coef*(image[i].b-backgnd[i].b))>>ShiftCount)+backgnd[i].b);

    if (BYTESPERPIXEL==3)
      *((unsigned *)(tmpdest))=Color24Bit(p.r, p.g, p.b);
    else
    if (BYTESPERPIXEL==4)
      *((unsigned *)(tmpdest))=Color32Bit(p.r, p.g, p.b);

    tmpdest+=BYTESPERPIXEL;
  }
}
//---------------------------------------------------------------------------
__inline void GrasomusAlphaBlend::AlphaBlendSoftwareMMX(PixelInfo *image, PixelInfo *backgnd, unsigned *rendered_image, unsigned char src_coef, unsigned char ShiftCount)
{
  unsigned PACK256 = 0x01000100;
  register unsigned PACKIMAGE, PACKBACKGND;
  unsigned PACKCOEF = src_coef | (src_coef<<16);

  unsigned PACKR, PACKG, PACKB;

  BYTE *tmpdest=(BYTE *)rendered_image;

  for (unsigned i=0; i<NO_OF_PIXELS; i+=2)
  {
    // render red channel
    // pack 2 pixels of image
    PACKIMAGE = image[i].r | (image[i+1].r<<16);
    // pack 2 pixels of background
    PACKBACKGND = backgnd[i].r | (backgnd[i+1].r<<16);

    // pack 2 word into a 32-bit register
    PACKR=(((((PACKIMAGE+PACK256-PACKBACKGND)*src_coef)>>8)+PACKBACKGND-PACKCOEF)>>3)&0x001F001F;

    // render green channel
    // pack 2 pixels of image
    PACKIMAGE = image[i].g | (image[i+1].g<<16);
    // pack 2 pixels of background
    PACKBACKGND = backgnd[i].g | (backgnd[i+1].g<<16);

    // pack 2 word into a 32-bit register
    PACKG=(((((PACKIMAGE+PACK256-PACKBACKGND)*src_coef)>>8)+PACKBACKGND-PACKCOEF)>>3)&0x001F001F;

    // render blue channel
    // pack 2 pixels of image
    PACKIMAGE = image[i].b | (image[i+1].b<<16);
    // pack 2 pixels of background
    PACKBACKGND = backgnd[i].b | (backgnd[i+1].b<<16);

    // pack 2 word into a 32-bit register
    PACKB=(((((PACKIMAGE+PACK256-PACKBACKGND)*src_coef)>>8)+PACKBACKGND-PACKCOEF)>>3)&0x001F001F;

    *((unsigned *)(tmpdest))=(PACKR<<g_RShift) | (PACKG<<g_GShift) | PACKB;
    tmpdest+=4;
    
  }
}
//---------------------------------------------------------------------------
void GrasomusAlphaBlend::AlphaBlendHardwareMMX(PixelInfo *image, PixelInfo *backgnd, unsigned *rendered_image, unsigned char src_coef, unsigned char ShiftCount)
{
  WORD packimage[4];
  WORD packbackgnd[4];

  BYTE *timage=(BYTE *)packimage;
  BYTE *tbackgnd=(BYTE *)packbackgnd;

  __int64 RShift=g_RShift;
  __int64 GShift=g_GShift;

  DWORD ALPHA=(DWORD)src_coef;
  __int64 PACK256=0x0100010001000100;
  __int64 PACKALPHA, PACKR, PACKG, PACKB;
  __int64 GLOBALMASK=0x001F001F001F001F;

  BYTE *tmpdest=(BYTE *)rendered_image;

  // pack ALPHA
  asm
  {
    movd mm2,ALPHA     // Copy ALPHA into mm2
    punpcklwd mm2,mm2  // Unpack mm2 - 0000 0000 00aa 00aa
    punpckldq mm2,mm2  // Unpack mm2 - 00aa 00aa 00aa 00aa
    movq PACKALPHA,mm2 // Save the result into PACKALPHA
  }

  for (unsigned i=0; i<NO_OF_PIXELS; i+=4)
  {
    // render red channel
    // pack 4 pixels of image
    packimage[0]=image[i].r;
    packimage[1]=image[i+1].r;
    packimage[2]=image[i+2].r;
    packimage[3]=image[i+3].r;
    packbackgnd[0]=backgnd[i].r;
    packbackgnd[1]=backgnd[i+1].r;
    packbackgnd[2]=backgnd[i+2].r;
    packbackgnd[3]=backgnd[i+3].r;
    asm
    {
      push edi              // use this to point to tmpimage
      push esi              // and this to point to tmpbackgnd
      // render red channel
      
      mov edi, timage
      mov esi, tbackgnd
      movq mm0, [edi]       // Copy image into mm0
      movq mm1, [esi]       // Copy backgnd into mm1
      paddw mm0, PACK256    // PACKIMAGE + PACK256 - PACKBACKGND
      psubw mm0, mm1
      pmullw mm0, PACKALPHA // result * ALPHA, then >> 8
      psrlw mm0, 8
      paddw mm0, [esi]      // result + packbackgnd - alpha
      psubw mm0, PACKALPHA
      psrlw mm0, 3          // shift result right by 3, then and with 0x1F
      pand mm0, GLOBALMASK
      movq PACKR, mm0       // store result to PACKR
      pop esi
      pop edi
    }

    // render green channel
    // pack 4 pixels of image
    packimage[0]=image[i].g;
    packimage[1]=image[i+1].g;
    packimage[2]=image[i+2].g;
    packimage[3]=image[i+3].g;
    packbackgnd[0]=backgnd[i].g;
    packbackgnd[1]=backgnd[i+1].g;
    packbackgnd[2]=backgnd[i+2].g;
    packbackgnd[3]=backgnd[i+3].g;
    asm
    {
      push edi              // use this to point to tmpimage
      push esi              // and this to point to tmpbackgnd
      mov edi, timage
      mov esi, tbackgnd
      movq mm0, [edi]       // Copy image into mm0
      movq mm1, [esi]       // Copy backgnd into mm1
      paddw mm0, PACK256    // PACKIMAGE + PACK256 - PACKBACKGND
      psubw mm0, mm1
      pmullw mm0, PACKALPHA // result * ALPHA, then >> 8
      psrlw mm0, 8
      paddw mm0, [esi]      // result + packbackgnd - alpha
      psubw mm0, PACKALPHA
      psrlw mm0, 3          // shift result right by 3, then and with 0x1F
      pand mm0, GLOBALMASK
      movq PACKG, mm0       // store result to PACKG
      pop esi
      pop edi
    }

    // render blue channel
    // pack 4 pixels of image
    packimage[0]=image[i].b;
    packimage[1]=image[i+1].b;
    packimage[2]=image[i+2].b;
    packimage[3]=image[i+3].b;
    packbackgnd[0]=backgnd[i].b;
    packbackgnd[1]=backgnd[i+1].b;
    packbackgnd[2]=backgnd[i+2].b;
    packbackgnd[3]=backgnd[i+3].b;
    asm
    {
      push edi              // use this to point to tmpimage
      push esi              // and this to point to tmpbackgnd
      mov edi, timage
      mov esi, tbackgnd
      movq mm0, [edi]       // Copy image into mm0
      movq mm1, [esi]       // Copy backgnd into mm1
      paddw mm0, PACK256    // PACKIMAGE + PACK256 - PACKBACKGND
      psubw mm0, mm1
      pmullw mm0, PACKALPHA // result * ALPHA, then >> 8
      psrlw mm0, 8
      paddw mm0, [esi]      // result + packbackgnd - alpha
      psubw mm0, PACKALPHA
      psrlw mm0, 3          // shift result right by 3, then and with 0x1F
      pand mm0, GLOBALMASK
      movq PACKB, mm0       // store result to PACKB
      pop esi
      pop edi
    }

    // now we'd have to transfer it back to the memory location pointed by tmpdest
    asm
    {
      movq mm0, PACKR
      psllw mm0, RShift
      movq mm1, PACKG
      psllw mm1, GShift
      movq mm2, PACKB
      por mm0, mm1
      por mm0, mm2
      push edi
      mov edi, tmpdest
      movq [edi], mm0
      pop edi
    }

    tmpdest+=8;

  }
  // clean up MMX registers
  asm
  {
    emms
  }
}
//---------------------------------------------------------------------------
void __fastcall TAlphaBlenderForm::FormHide(TObject *Sender)
{
  unsigned start_cnt=GetTickCount();
  unsigned end_cnt;

  do
  {
    end_cnt=GetTickCount();
  }while(end_cnt-start_cnt<500);
  start_cnt=end_cnt;

  unsigned delay=Delay;

  unsigned short StartF=(unsigned short)(255-MaxOpacity);
  unsigned short EndF=(unsigned short)(255-MinOpacity);
  for (unsigned short i=StartF; i<EndF;)
  {
    if (BITSPERPIXEL==16)
    {
      if (g_isMMX)
        AlphaB.AlphaBlendHardwareMMX(backgnd, image, frames, i, ShiftCount);
      else
        AlphaB.AlphaBlendSoftwareMMX(backgnd, image, frames, i, ShiftCount);
    }
    else
      AlphaB.AlphaBlendBasic(backgnd, image, frames, i, ShiftCount);
    SetBitmapBits(hbm, NO_OF_PIXELS*AlphaB.BYTESPERPIXEL, frames);
    BitBlt(AlphaBlenderFormDC, 0, 0, bm.bmWidth, bm.bmHeight, hMemoryDC, 0, 0, SRCCOPY);

    do
    {
      end_cnt=GetTickCount();
    }while(end_cnt-start_cnt<delay);

    i+=(unsigned short)((end_cnt-start_cnt)/delay);
    start_cnt=end_cnt;
  }

  SelectObject(hMemoryDC, hbmOld);
  DeleteObject(hbm);
  DeleteDC(hMemoryDC);
  ReleaseDC(hDesktopDC, NULL);
  ReleaseDC(AlphaBlenderFormDC, Handle);
  SystemParametersInfo( SPI_SCREENSAVERRUNNING, 0, NULL, 0);
  ClipCursor(NULL);
  ::SetCursor(oldcursor);

}
//---------------------------------------------------------------------------
